from migen import *
from migen.genlib.cdc import PulseSynchronizer, MultiReg
from misoc.interconnect.csr import *


class DDMTDSampler(Module):
    def __init__(self, cd_ref, main_clk_se):
        self.ref_beating = Signal()
        self.main_beating = Signal()

        # # #

        ref_beating_FF = Signal()
        main_beating_FF = Signal()
        self.specials += [
            # Two back to back FFs are used to prevent metastability
            Instance("FD", i_C=ClockSignal("helper"),
                     i_D=cd_ref.clk, o_Q=ref_beating_FF),
            Instance("FD", i_C=ClockSignal("helper"),
                     i_D=ref_beating_FF, o_Q=self.ref_beating),
            Instance("FD", i_C=ClockSignal("helper"),
                     i_D=main_clk_se, o_Q=main_beating_FF),
            Instance("FD", i_C=ClockSignal("helper"),
                     i_D=main_beating_FF, o_Q=self.main_beating)
        ]


class DDMTDDeglitcherMedianEdge(Module):
    def __init__(self, counter, input_signal, stable_0_period=100, stable_1_period=100):
        self.tag = Signal(len(counter))
        self.detect = Signal()

        stable_0_counter = Signal(reset=stable_0_period - 1, max=stable_0_period)
        stable_1_counter = Signal(reset=stable_1_period - 1, max=stable_1_period)

        # # #

        # Based on CERN's median edge deglitcher FSM
        # https://white-rabbit.web.cern.ch/documents/Precise_time_and_frequency_transfer_in_a_White_Rabbit_network.pdf (p.72)
        fsm = ClockDomainsRenamer("helper")(FSM(reset_state="WAIT_STABLE_0"))
        self.submodules += fsm

        fsm.act("WAIT_STABLE_0",
            If(stable_0_counter != 0,
                NextValue(stable_0_counter, stable_0_counter - 1)
            ).Else(
                NextValue(stable_0_counter, stable_0_period - 1),
                NextState("WAIT_EDGE")
            ),
            If(input_signal,
                NextValue(stable_0_counter, stable_0_period - 1)
            ),
        )
        fsm.act("WAIT_EDGE",
            If(input_signal,
                NextValue(self.tag, counter),
                NextState("GOT_EDGE")
            )
        )
        fsm.act("GOT_EDGE",
            If(stable_1_counter != 0,
                NextValue(stable_1_counter, stable_1_counter - 1)
            ).Else(
                NextValue(stable_1_counter, stable_1_period - 1),
                self.detect.eq(1),
                NextState("WAIT_STABLE_0")
            ),
            If(~input_signal,
                NextValue(self.tag, self.tag + 1),
                NextValue(stable_1_counter, stable_1_period - 1)
            ),
        )
        

class DDMTD(Module):
    def __init__(self, counter, input_signal):

        # in helper clock domain
        self.h_tag = Signal(len(counter))
        self.h_tag_update = Signal()

        # # #

        deglitcher = DDMTDDeglitcherMedianEdge(counter, input_signal)
        self.submodules += deglitcher

        self.sync.helper += [
            self.h_tag_update.eq(0),
            If(deglitcher.detect,
                self.h_tag_update.eq(1),
                self.h_tag.eq(deglitcher.tag)
               )
        ]